/******************************************************************************
 *   Copyright (C) 2006 by Igor Timoshenko  (timigor@yandex.ru).              *
 *  Use may be in whole or in part in accordance to the                       *
 *  General Public License (GPL).                                             *
*   Simple net driver - transmiter                                            *
 ******************************************************************************/
#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/init.h>
#include <linux/netdevice.h>
#include <linux/skbuff.h>
#include <asm/io.h> 	/* -   -*/
#include <linux/jiffies.h>  /*    */
#include <linux/timer.h>

#define BASE     0x3f8      /*   COM1 \COM2 = 0x2f8\   */
#define ICR_N     1         /*   */
#define IIDR_N    2         /*    / . */
#define MCR_N     4         /*   */
#define LSR_N     5         /*   */
#define MSR_N     6         /*   */

/*   */
typedef union  _MCR_ {
   struct {
      unsigned char dtr : 1;
      unsigned char rts : 1;
      unsigned char other : 6;
   } bit_reg;
   unsigned char byte;
} MCR;
MCR reg_mcr;
/*   */
typedef union _MSR_ {
   struct {
      unsigned char other1 : 4;
      unsigned char cts : 1;
      unsigned char dsr : 1;
      unsigned char other2 : 2;
   } bit_reg;
   unsigned char byte;
} MSR;
MSR reg_msr;
/*   */
typedef union _LSR_ {
   struct {
      unsigned char in_ready : 1;
      unsigned char other1 : 4;
      unsigned char out_ready : 1;
      unsigned char other2 : 2;
   } bit_reg;
   unsigned char byte;
} LSR;
LSR reg_lsr;
/*  IP */
struct iphdr  {
    unsigned int ihl:4;
    unsigned int version:4;
    u_int8_t tos;
    u_int16_t tot_len;
    u_int16_t id;
    u_int16_t frag_off;
    u_int8_t ttl;
    u_int8_t protocol;
    u_int16_t check;
    unsigned char saddr[4]; /* IP address of originator */
    unsigned char daddr[4]; /* IP address of destination */
    /*The options start here. */
  };
/*   */
struct my_buf {
int tlen;
	union tbuffer{
	unsigned char tail[2];	/* */
	unsigned char tbuff[1024];
	struct iphdr thdr;
	}tb;
};
struct net_device_stats *stats;
struct my_buf xbf;
int retval, shift, i;

int ssl_stop (struct net_device *dev);
int ssl_open (struct net_device *dev);
static int ssl_xmit(struct sk_buff *skb, struct net_device *dev);
//static void load_pack(struct my_buf *xbf);
//void tpdumpk(struct my_buf *xbf);
unsigned char *sad;
unsigned char *dad;
unsigned char bufaddr[4]; /* IP address*/
/*  */  
struct net_device ssl_dev = {
	.name			= "ssl",
	.open			= ssl_open,
	.stop			= ssl_stop,
	.hard_start_xmit	= ssl_xmit,
};
static struct net_device_stats *ssl_get_stats(struct net_device *dev)
{ return dev->priv; }
/* ######################################################*/
static int ssl_xmit (struct sk_buff *skb, struct net_device *dev)
{
	if (skb->len > 1024 - 3) {
	printk(KERN_WARNING "ssl: T_buffer is small, dropping packet.\n");
	stats->tx_dropped++;
	} else {
		/*    */
		shift = 0;
		while (shift + 1 <= skb->len) {
			while ((inb_p(BASE + LSR_N) & 0x20)==0);
			outb_p(skb->data[shift], BASE);
			shift++;
			}
	printk(KERN_WARNING "ssl: Transmit to COM-port.\n");
	printk("\n");
	dev_kfree_skb(skb);
	stats->tx_bytes += skb->len;
	stats->tx_packets++;
	}
return 0;
}
/*=================== =========================*/
/*static void load_pack(struct my_buf *xbf)
{
	struct net_device_stats *stats = ssl_dev.priv;
	struct sk_buff *r_skb;
	//printk(KERN_WARNING "ssl:  .\n");
	r_skb = dev_alloc_skb(xbf->tlen);
		if (r_skb == NULL)  {
			printk(KERN_WARNING "ssl: memory squeeze, dropping packet.\n");
			stats->rx_dropped++;
			return;
		}
	r_skb->dev = &ssl_dev;
	memcpy(skb_put(r_skb, xbf->tlen), xbf->tb.tbuff, xbf->tlen);
	//r_skb->mac.raw=r_skb->data;
	r_skb->protocol=htons(ETH_P_IP);
	
	//rpdumpk(xbf);   /*    
	netif_rx(r_skb);
	stats->rx_bytes += xbf->tlen;
	stats->rx_packets++;
}*/
/*  ###########################*/
int ssl_open (struct net_device *dev)
{
printk(KERN_WARNING "ssl: ssl_open called.\n");
netif_start_queue (dev);
/* COM1 */
outb_p(0, BASE + IIDR_N);	/* */
outb_p(0, BASE + ICR_N);	/* */
reg_mcr.byte = 0;		/*   */

return 0;
}
/*  #############################*/
int ssl_stop (struct net_device *dev)
{
printk (KERN_WARNING "ssl: ssl_stop called.\n");
netif_stop_queue(dev);
	return 0;
}
/*============= ===============*/
int ssl_init_module (void)
{
stats = kmalloc(sizeof(struct net_device_stats), GFP_KERNEL);
	if (stats) {
	memset(stats, 0, sizeof(struct net_device_stats));
	ssl_dev.priv = stats;
	ssl_dev.get_stats = &ssl_get_stats;
	}
	if ((retval = register_netdev (&ssl_dev))) {
	printk (KERN_WARNING "ssl: Error %d  of initializing module.\n",retval);
	return retval;
	}
printk(KERN_WARNING "ssl: initializing the module.\n");
return 0;
}
/*============ =======================*/
void ssl_cleanup (void)
{
unregister_netdev (&ssl_dev);
printk (KERN_WARNING "ssl: Cleaning Up the Module.\n");
return;
}
module_init (ssl_init_module);
module_exit (ssl_cleanup);
/*  ###################################### */
/*       .*/
/*void printAddr(unsigned char *addr)
{
int j;
	for ( j = 0; j < 4; j++ )
	{
	printk("%d", addr[j]);
	if ( j < 3 )
		printk(".");
	}
}*/
/*   */
/*void tpdumpk(struct my_buf *txbf)
{
int i;
	printk("\n---RETURNING---");
	for ( i = 0; i < txbf->tlen; i++ )
	{
		if ( !(i & 15) ) printk("\n%04X:  ", i);
		printk("%02X ", ((unsigned char*)txbf->tb.tbuff)[i]);
	}
	printk("\n");
	printk("IPv%d: hdr-size=%d pkt-size=%d protocol=%d TTL=%d ",
		txbf->tb.thdr.version, txbf->tb.thdr.ihl*4, ntohs(txbf->tb.thdr.tot_len),
		txbf->tb.thdr.protocol, txbf->tb.thdr.ttl);
printk("\n");
	printk("rst= ");
printAddr(txbf->tb.thdr.saddr);
	printk(" dst= ");
printAddr(txbf->tb.thdr.daddr);
printk("\n");
}*/
