/******************************************************************************
 *   Copyright (C) 2006 by Igor Timoshenko  (timigor@yandex.ru).              *
 *  Use may be in whole or in part in accordance to the                       *
 *  General Public License (GPL).                                             *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND    *
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE      *
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE *
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE   *
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL *
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS    *
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)      *
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT *
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY  *
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF     *
 * SUCH DAMAGE.                                                               *
 *                                                                            *
 *   Simple SLIP     ver. 1.1                                                 *
 ******************************************************************************/
#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/init.h>
#include <linux/netdevice.h>
#include <linux/skbuff.h>
#include <asm/io.h> 		/* -   -*/
#include <linux/jiffies.h>  	/*    */
#include <linux/timer.h>

#define T_TICK     HZ/100

#define BASE     0x3f8      /*   COM1 \COM2 = 0x2f8\   */
#define ICR_N     1         /*   */
#define IIDR_N    2         /*    / . */
#define MCR_N     4         /*   */
#define LSR_N     5         /*   */
#define MSR_N     6         /*   */

/*   */
typedef union  _MCR_ {
   struct {
      unsigned char dtr : 1;
      unsigned char rts : 1;
      unsigned char other : 6;
   } bit_reg;
   unsigned char byte;
} MCR;
MCR reg_mcr;
/*   */
typedef union _MSR_ {
   struct {
      unsigned char other1 : 4;
      unsigned char cts : 1;
      unsigned char dsr : 1;
      unsigned char other2 : 2;
   } bit_reg;
   unsigned char byte;
} MSR;
MSR reg_msr;
/*   */
typedef union _LSR_ {
   struct {
      unsigned char in_ready : 1;
      unsigned char other1 : 4;
      unsigned char out_ready : 1;
      unsigned char other2 : 2;
   } bit_reg;
   unsigned char byte;
} LSR;
LSR reg_lsr;
/*  IP */
struct iphdr  {
    unsigned int ihl:4;
    unsigned int version:4;
    u_int8_t tos;
    u_int16_t tot_len;
    u_int16_t id;
    u_int16_t frag_off;
    u_int8_t ttl;
    u_int8_t protocol;
    u_int16_t check;
    unsigned char saddr[4]; /* IP address of originator */
    unsigned char daddr[4]; /* IP address of destination */
    /*The options start here. */
  };
/*   */
struct my_buf {
int tlen;
unsigned char tail[2];
	union tbuffer{
	unsigned char tbuff[1024];
	struct iphdr thdr;
	}tb;
};
struct net_device_stats *stats;
struct my_buf xbf;

struct timer_list timer;
spinlock_t timer_lock = SPIN_LOCK_UNLOCKED;

unsigned long timeout;
int retval, shift, i;

int ssl_stop (struct net_device *dev);
int ssl_open (struct net_device *dev);
static int ssl_xmit(struct sk_buff *skb, struct net_device *dev);
static void load_pack(struct my_buf *xbf);
void tpdumpk(struct my_buf *xbf);
unsigned char *sad;
unsigned char *dad;
unsigned char bufaddr[4]; /* IP address*/
/*  */  
struct net_device ssl_dev = {
	.name			= "ssl",
	.open			= ssl_open,
	.stop			= ssl_stop,
	.hard_start_xmit	= ssl_xmit,
};
static struct net_device_stats *ssl_get_stats(struct net_device *dev)
{ return dev->priv; }

/* =######################################################*/
void ssl_irq (unsigned long data) {
spin_lock_bh(&timer_lock);

struct net_device_stats *stats = ssl_dev.priv;
struct my_buf *qbf = &xbf;
/*    ttyS0  RTS->  DSR<-*/
reg_msr.byte = inb_p(BASE + MSR_N);
if (reg_msr.bit_reg.dsr != 0) {		/*ģ   */
reg_mcr.bit_reg.rts = 1;
outb_p(reg_mcr.byte, BASE + MCR_N);	/*  ɣ*/
     shift = 0;
     while (reg_msr.bit_reg.dsr == 1) { /*  */
  	while ((inb_p(BASE + LSR_N) & 1)==0); /*ģ    ɣ*/
	qbf->tb.tbuff[++shift-2] = inb_p(BASE); /*  */
		if (shift > 1024) {
		printk(KERN_WARNING "ssl: R_buffer is small and overfull, dropping packet.\n");
		stats->rx_dropped++;
		goto r_out;
		}
     reg_msr.byte = inb_p(BASE + MSR_N);
     }	/* */
     	if (shift == 0) {
	printk(KERN_WARNING "ssl: Packet is 0 byte, dropping packet.\n");
	stats->rx_dropped++;
	goto r_out;
	}
     qbf->tlen = shift-2;
     tpdumpk(&xbf);	/*      */
     load_pack(&xbf);	/*    SKB*/
} //else printk(KERN_WARNING "ssl: Line DSR is free.\n");
r_out:
reg_mcr.bit_reg.rts = 0;
outb_p(reg_mcr.byte, BASE + MCR_N);  /*  ɣ*/

mod_timer(&timer, jiffies + T_TICK);
spin_unlock_bh(&timer_lock);
}
/*=================== =========================*/
static void load_pack(struct my_buf *xbf) {
	struct net_device_stats *stats = ssl_dev.priv;
	struct sk_buff *r_skb;
	r_skb = dev_alloc_skb(xbf->tlen);
		if (r_skb == NULL)  {
			printk(KERN_WARNING "ssl: memory squeeze, dropping packet.\n");
			stats->rx_dropped++;
			return;
		}
	r_skb->dev = &ssl_dev;
	memcpy(skb_put(r_skb, xbf->tlen), xbf->tb.tbuff, xbf->tlen);
	r_skb->protocol=htons(ETH_P_IP);
	
	netif_rx(r_skb);
	stats->rx_bytes += xbf->tlen;
	stats->rx_packets++;
}

/* ######################################################*/
static int ssl_xmit (struct sk_buff *skb, struct net_device *dev) {
	if (skb->len > 1024 - 3) {
	printk(KERN_WARNING "ssl: T_buffer is small, dropping packet.\n");
	stats->tx_dropped++;
	} else {
	/*    ttyS0  DTR->  CTS<-*/
	reg_lsr.byte = inb_p(BASE + LSR_N);
	if (reg_lsr.bit_reg.out_ready != 0) {
		reg_msr.byte = inb_p(BASE + MSR_N);	
			if (reg_msr.bit_reg.cts == 0) {
			reg_mcr.bit_reg.dtr = 1;
			outb_p(reg_mcr.byte, BASE + MCR_N);	/* */
			timeout = jiffies + T_TICK;
			while (reg_msr.bit_reg.cts == 0) {  	/*ģ  CTS  ɣ*/
			reg_msr.byte = inb_p(BASE + MSR_N);	/**/
				if (jiffies > timeout) {
				printk(KERN_WARNING "ssl: The CTS response is absent control time, dropping packet.\n");
				stats->tx_dropped++;
				goto t_out;
				}
			}
			/*    */
			shift = 0;
			while (shift + 1 <= skb->len) {
				while ((inb_p(BASE + LSR_N) & 0x20)==0);/*ģ    */
				outb_p(skb->data[shift], BASE); /*  */
				shift++;
			}
			while ((inb_p(BASE + LSR_N) & 0x20)==0);	/*Push bufer UART*/
			outb_p(0, BASE);
			while ((inb_p(BASE + LSR_N) & 0x20)==0);
			outb_p(0, BASE);
			} //else printk(KERN_WARNING "ssl: Line CTS is busy.\n");
		//} //else printk(KERN_WARNING "ssl: Transmit buf is empty.\n");
	} //else printk(KERN_WARNING "ssl: Out is full.\n");
	t_out:
	reg_mcr.bit_reg.dtr = 0;
	outb_p(reg_mcr.byte, BASE + MCR_N);	/* */
	
	dev_kfree_skb(skb);
	stats->tx_bytes += skb->len;
	stats->tx_packets++;
	}
return 0;
}

/*  ###########################*/
int ssl_open (struct net_device *dev) {
printk(KERN_WARNING "ssl: ssl_open called.\n");
netif_start_queue (dev);
/* COM1 */
outb_p(0, BASE + IIDR_N);	/* */
outb_p(0, BASE + ICR_N);	/* */
reg_mcr.byte = 0;		/*   */
/* */
init_timer(&timer);
timer.expires = jiffies + HZ; /*  .,   */
timer.function = ssl_irq;
add_timer(&timer);
return 0;
}
/*  #############################*/
int ssl_stop (struct net_device *dev) {
del_timer_sync(&timer);		/* */
printk (KERN_WARNING "ssl: ssl_stop called.\n");
netif_stop_queue(dev);
return 0;
}
/*============= ===============*/
int ssl_init_module (void) {
stats = kmalloc(sizeof(struct net_device_stats), GFP_KERNEL);
	if (stats) {
	memset(stats, 0, sizeof(struct net_device_stats));
	ssl_dev.priv = stats;
	ssl_dev.get_stats = &ssl_get_stats;
	}
	if ((retval = register_netdev (&ssl_dev))) {
	printk (KERN_WARNING "ssl: Error %d  of initializing module.\n",retval);
	return retval;
	}
printk(KERN_WARNING "ssl: initializing the module.\n");
return 0;
}
/*============ =======================*/
void ssl_cleanup (void) {
unregister_netdev (&ssl_dev);
printk (KERN_WARNING "ssl: Cleaning Up the Module.\n");
return;
}
module_init (ssl_init_module);
module_exit (ssl_cleanup);
